﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Threading.Tasks;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models.CustomProjects;
using ClientServer.Models.Users;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Controllers.Core.customProjects
{
    /// <summary>
    /// custom projects overview controller
    /// </summary>
    [Route(Constants.ApiPrefix + "custom/projects/overview")]
    public class CustomProjectsOverviewController : ControllerWithDb
    {
        public CustomProjectsOverviewController(SyndromDbContext context) : base(context)
        {
        }

        /// <summary>
        /// returns all custom projects of the current user
        /// </summary>
        [HttpPost("all")]
        public async Task GetOwnCustomProjects([FromBody] PaginationPostData pagination)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProjects = _context.CustomProjects
                    .Where(p => p.UserId == userId)
                ;

            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                customProjects = customProjects
                    .Where(p => (p.DisplayName != null && p.DisplayName.ToLower().Contains(text))
                    );
            }

            string orderByProp = "";
            Expression<Func<CustomProject, object>> orderByFunc = null;

            //not all props are sortable via the frontend
            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "Id".ToLower())
                {
                    orderByFunc = p => p.Id;
                    orderByProp = "Id".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "DisplayName".ToLower())
                {
                    orderByFunc = p => p.DisplayName;
                    orderByProp = "DisplayName".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "LastEditorPLangId".ToLower())
                {
                    orderByFunc = p => p.LastEditorPLangId;
                    orderByProp = "LastEditorPLangId".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "LastUpdatedAt".ToLower())
                {
                    orderByFunc = p => p.LastUpdatedAt;
                    orderByProp = "LastUpdatedAt".ToLower();
                }
            }

            var paginatedData = await PaginationHelper.GetAsPaginationAsync(customProjects,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );

            var customProjectOverviews = paginatedData.Items.Select(p => new CustomProjectPreviewFullBase()
            {
                Id = p.Id,
                DisplayName = p.DisplayName,
                LastEditorPLangId = p.LastEditorPLangId,
                LastUpdatedAt = p.LastUpdatedAt
            }).ToList();

            var frontendData = PaginationHelper.TransformForFrontend(paginatedData, customProjectOverviews);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<CustomProjectPreviewFullBase>>(ResponseCode.Ok,
                        "",
                        frontendData)));
        }

        /// <summary>
        /// tries to delete the given custom project via id of the current user
        /// </summary>
        /// <param name="customProjectId"></param>
        [HttpDelete("delete/{customProjectId}")]
        public async Task DeleteOwnCustomProjects(int customProjectId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var customProject = await _context.CustomProjects
                    .FirstOrDefaultAsync(p => p.UserId == userId && p.Id == customProjectId)
                ;

            if (customProject == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "custom project not found")));
                return;
            }

            try
            {
                _context.CustomProjects.Remove(customProject);
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }

            await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "custom project deleted")));
        }
    }


    public class CustomProjectPreviewFullBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the last edited plang... set through the custom solution
        /// </summary>
        public int LastEditorPLangId { get; set; }

        public DateTime LastUpdatedAt { get; set; }
    }
}
