﻿using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using ClientServer.Controllers.Core.Testing;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models.Exercises.Release;
using ClientServer.Models.Exercises.Solution;
using ClientServer.Models.Exercises.Tests;
using ClientServer.Schedulers;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Primitives;
using Newtonsoft.Json;
using CsvHelper;
using CsvHelper.Configuration;
using ClientServer.Models;

namespace ClientServer.Controllers.Core.Exercises
{
    [Route(Constants.ApiPrefix + "submissions")]
    public class SubmissionController : ControllerWithDb
    {
        /// <summary>
        /// utf8 encoding without bom
        /// </summary>
        public static Encoding UTF8WithoutBom = new UTF8Encoding(false);

        public readonly static char CsvCommentChar = '#';
        public readonly static string CsvDelimiterChar = ",";

        /// <summary>
        /// the name of the dir for the code template
        /// we include the code templeta for download all submissions so one can compare if the user has changed anything...
        /// </summary>
        public readonly static string DefaultSubmissionWithCodeTemplatesDirName = "template";

        public SubmissionController(SyndromDbContext context) : base(context)
        {
        }


        /// <summary>
        /// gets the submissions (results) for a release (and the release itself)
        /// </summary>
        /// <param name="releaseId"></param>
        /// <param name="getPreAssessmentResults">
        /// null: assessments or pre assessments is determined by the release (isReleased) itself
        /// true: return the results of the normal test attached to the latest user solution (version)
        ///         (the results of the normal test from the user solution are taken, submissions tests are ignored)
        /// false: return the assessment results attached to the solution (results only contain values if the assessment ran)
        ///
        /// </param>
        [HttpPost("get/{releaseId}/{getPreAssessmentResults?}")]
        public async Task GetSubmissionResults([FromBody] PaginationPostData pagination, int releaseId,
            bool? getPreAssessmentResults = null)
        {
            if (!await base.IsLoggedIn()) return;

            //get all automatic assessments for this release

            //first check if released
            var release = await _context.ExerciseReleases
                    .FirstOrDefaultAsync(p => p.Id == releaseId)
                ;


            bool shouldGetPreAssessmentResults = false;

            if (getPreAssessmentResults == null)
            {
                shouldGetPreAssessmentResults = release.IsReleased;
            }
            else if (getPreAssessmentResults == true)
            {
                shouldGetPreAssessmentResults = true;
            }
            else
            {
                shouldGetPreAssessmentResults = false;
            }


            //then query data

            if (shouldGetPreAssessmentResults)
            {
                //we also need the solution version with the normal test results

                release = await _context.ExerciseReleases
                        .Include(p => p.Exercise)
                        .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                        .ThenInclude(p => p.User)
                        .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                        .ThenInclude(p => p.Solutions)
                        .ThenInclude(p => p.Assessment)
                        .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                        .ThenInclude(p => p.Solutions)
                        .ThenInclude(p => p.TestResults)
                        .FirstOrDefaultAsync(p => p.Id == releaseId)
                    ;
            }
            else
            {
                release = await _context.ExerciseReleases
                        .Include(p => p.Exercise)
                        .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                        .ThenInclude(p => p.User)
                        .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                        .ThenInclude(p => p.Solutions)
                        .ThenInclude(p => p.Assessment)
                        .FirstOrDefaultAsync(p => p.Id == releaseId)
                    ;
            }


            if (release == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }


            //the user group that manages the exercise
            var targetUserGroup =
                await _context.ExerciseReleases.Where(p => p.Id == release.Id)
                    .Select(p => p.Exercise.UserGroupId)
                    .FirstOrDefaultAsync();

            if (!await base.HasGroupPermission(targetUserGroup,
                permission => permission != null && permission.CanAssessExercises))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            PaginatedData<AssessmentFullBase> assessments;

            if (shouldGetPreAssessmentResults)
            {
                var normalTests = await _context.Tests
                        .Where(p => p.ExerciseId == release.ExerciseId && p.IsSubmitTest == false)
                        .ToListAsync()
                    ;

                assessments = await this.GetPreAssessmentResults(release, normalTests, pagination);
            }
            else
            {
                assessments = await this.GetAssessmentResults(release, pagination);
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<AssessmentFullBase>>(ResponseCode.Ok,
                        "assessments queried", assessments)));
        }
        
        /// <summary>
        /// 
        /// </summary>
        /// <param name="releaseCode"></param>
        /// <param name="sortKey">see the frontend columns for possible values</param>
        /// <param name="ascending"></param>
        [HttpPost("export/csv/{releaseCode}/{sortKey}/{IsOrderByDesc}/{csvDelimiter}")]
        public async Task GetResultsAsCsv(string releaseCode, string sortKey, bool IsOrderByDesc, string csvDelimiter, [FromBody] CsvExportAssessmentOptionsObj csvExportAssessmentOptionsObj)
        {
            // because we use a link we don't get and need the csrf token...
            if (!await base.IsLoggedIn(null, false, true)) return;

            if (csvDelimiter.Trim() == "") {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "delimiter was empty or just whitespaces")));
                return;
            }

            csvDelimiter = csvDelimiter.Trim();

            sortKey = sortKey.ToLower();
            //see src/components/sites/manageOwnOrGroupExerciseComponents/submissionsSite/listView.tsx
            //for keys
            var possibleSortKeys = new Dictionary<string, Func<AssessmentFullBase, object>>()
            {
                ["token"] = p => p.Token,
                ["lastname"] = p => p.LastName,
                ["firstname"] = p => p.FirstName,
                ["plangid"] = p => p.PLangId,
                ["shouldnotcount"] = p => p.ShouldNotCount,
                ["participationcount"] = p => p.ParticipationCount,
                ["normaltestpoints"] = p => p.NormalTestPoints,
                ["maxNormalTestPoints"] = p => p.MaxNormalTestPoints,
                ["submittestpoints"] = p => p.SubmitTestPoints,
                ["maxSubmitTestPoints"] = p => p.MaxSubmitTestPoints,
                ["manualpoints"] = p => p.ManualPoints,
                ["MaxManualPoints".ToLower()] = p => p.MaxManualPoints,
            };

            if (possibleSortKeys.Keys.Contains(sortKey) == false)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                            "invalid sort key, possible ones: " + string.Join(", ", possibleSortKeys))));
                return;
            }

            var sortFunc = possibleSortKeys[sortKey];


            // var columnNames = new List<string>()
            // {
            //     "lastName",
            //     "firstName",
            //     "pLangId",
            //     "shouldNotCount",
            //     "participationCount",
            //     "normalTestPoints",
            //     "submitTestPoints",
            //     "manualPoints",
            //     "noteForOtherTutors",
            //     "FeedbackForStudent",
            //     "LastAssessmentErrorMessage"
            // };


            //just a temp array to store all needed info
            //                        display index, PropName, display col name
            var displayIndices = new List<Tuple<int, string, string>>();

            //we iterate all known cols/props then use this to get the target index in the result string array (csv row)
            //                                      prop name, display index
            var displayIndicesDict = new Dictionary<string, int>();

            //also check if all replace rules (regex) are valid
            
            #region all possible known column (even if not visible on frontend)
            
            //TODO maybe use array of props?? [csvExportAssessmentOptionsObj.Options.Token, ...] then loop?
                        
            if (csvExportAssessmentOptionsObj.Options.Token.willExport && csvExportAssessmentOptionsObj.Options.Token.displayIndex.HasValue) {
                
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.Token.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.Token), "Token"));
                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.Token), csvExportAssessmentOptionsObj.Options.Token.displayIndex.Value);

                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.Token,
                    nameof(csvExportAssessmentOptionsObj.Options.Token));

                if (message != null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }    
            }
            
            if (csvExportAssessmentOptionsObj.Options.FirstName.willExport  && csvExportAssessmentOptionsObj.Options.FirstName.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.FirstName.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.FirstName), "FirstName"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.FirstName), csvExportAssessmentOptionsObj.Options.FirstName.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.FirstName,
                    nameof(csvExportAssessmentOptionsObj.Options.FirstName));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.LastName.willExport  && csvExportAssessmentOptionsObj.Options.LastName.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.LastName.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.LastName), "LastName"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.LastName), csvExportAssessmentOptionsObj.Options.LastName.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.LastName,
                    nameof(csvExportAssessmentOptionsObj.Options.LastName));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.ShouldNotCount.willExport  && csvExportAssessmentOptionsObj.Options.ShouldNotCount.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.ShouldNotCount.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.ShouldNotCount), "ShouldNotCount"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.ShouldNotCount), csvExportAssessmentOptionsObj.Options.ShouldNotCount.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.ShouldNotCount,
                    nameof(csvExportAssessmentOptionsObj.Options.ShouldNotCount));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.UserId.willExport && csvExportAssessmentOptionsObj.Options.UserId.displayIndex.HasValue) {
                
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.UserId.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.UserId), "UserId"));
                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.UserId), csvExportAssessmentOptionsObj.Options.UserId.displayIndex.Value);

                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.UserId,
                    nameof(csvExportAssessmentOptionsObj.Options.UserId));

                if (message != null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }    
            }

            if (csvExportAssessmentOptionsObj.Options.PLangId.willExport  && csvExportAssessmentOptionsObj.Options.PLangId.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.PLangId.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.PLangId), "Programming Language"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.PLangId), csvExportAssessmentOptionsObj.Options.PLangId.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.PLangId,
                    nameof(csvExportAssessmentOptionsObj.Options.PLangId));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.ExerciseReleaseId.willExport && csvExportAssessmentOptionsObj.Options.ExerciseReleaseId.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.ExerciseReleaseId.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.ExerciseReleaseId), "ExerciseReleaseId"));
                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.ExerciseReleaseId), csvExportAssessmentOptionsObj.Options.ExerciseReleaseId.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.ExerciseReleaseId,
                    nameof(csvExportAssessmentOptionsObj.Options.ExerciseReleaseId));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.ParticipationCount.willExport  && csvExportAssessmentOptionsObj.Options.ParticipationCount.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.ParticipationCount.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.ParticipationCount), "ParticipationCount"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.ParticipationCount), csvExportAssessmentOptionsObj.Options.ParticipationCount.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.ParticipationCount,
                    nameof(csvExportAssessmentOptionsObj.Options.ParticipationCount));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
             
            if (csvExportAssessmentOptionsObj.Options.SubmitTestPoints.willExport  && csvExportAssessmentOptionsObj.Options.SubmitTestPoints.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.SubmitTestPoints.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.SubmitTestPoints), "SubmitTestPoints"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.SubmitTestPoints), csvExportAssessmentOptionsObj.Options.SubmitTestPoints.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.SubmitTestPoints,
                    nameof(csvExportAssessmentOptionsObj.Options.SubmitTestPoints));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints.willExport  && csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints), "MaxSubmitTestPoints"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints), csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints,
                    nameof(csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.NormalTestPoints.willExport  && csvExportAssessmentOptionsObj.Options.NormalTestPoints.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.NormalTestPoints.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.NormalTestPoints), "NormalTestPoints"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.NormalTestPoints), csvExportAssessmentOptionsObj.Options.NormalTestPoints.displayIndex.Value);
                
                                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.NormalTestPoints,
                    nameof(csvExportAssessmentOptionsObj.Options.NormalTestPoints));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints.willExport  && csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints), "MaxNormalTestPoints"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints), csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints,
                    nameof(csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.ManualPoints.willExport  && csvExportAssessmentOptionsObj.Options.ManualPoints.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.ManualPoints.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.ManualPoints), "ManualPoints"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.ManualPoints), csvExportAssessmentOptionsObj.Options.ManualPoints.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.ManualPoints,
                    nameof(csvExportAssessmentOptionsObj.Options.ManualPoints));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.MaxManualPoints.willExport  && csvExportAssessmentOptionsObj.Options.MaxManualPoints.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.MaxManualPoints.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.MaxManualPoints), "MaxManualPoints"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.MaxManualPoints), csvExportAssessmentOptionsObj.Options.MaxManualPoints.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.MaxManualPoints,
                    nameof(csvExportAssessmentOptionsObj.Options.MaxManualPoints));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.NoteForOtherTutors.willExport  && csvExportAssessmentOptionsObj.Options.NoteForOtherTutors.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.NoteForOtherTutors.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.NoteForOtherTutors), "NoteForOtherTutors"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.NoteForOtherTutors), csvExportAssessmentOptionsObj.Options.NoteForOtherTutors.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.NoteForOtherTutors,
                    nameof(csvExportAssessmentOptionsObj.Options.NoteForOtherTutors));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.FeedbackForStudent.willExport  && csvExportAssessmentOptionsObj.Options.FeedbackForStudent.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.FeedbackForStudent.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.FeedbackForStudent), "FeedbackForStudent"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.FeedbackForStudent),csvExportAssessmentOptionsObj.Options.FeedbackForStudent.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.FeedbackForStudent,
                    nameof(csvExportAssessmentOptionsObj.Options.FeedbackForStudent));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            
            
            if (csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage.willExport  && csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage), "LastAssessmentErrorMessage"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage), csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage,
                    nameof(csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
            
            if (csvExportAssessmentOptionsObj.Options.LastEditingIpAddress.willExport  && csvExportAssessmentOptionsObj.Options.LastEditingIpAddress.displayIndex.HasValue) {
                displayIndices.Add(new Tuple<int, string, string>(csvExportAssessmentOptionsObj.Options.LastEditingIpAddress.displayIndex.Value, nameof(csvExportAssessmentOptionsObj.Options.LastEditingIpAddress), "LastEditingIpAddress"));

                displayIndicesDict.Add(nameof(csvExportAssessmentOptionsObj.Options.LastEditingIpAddress), csvExportAssessmentOptionsObj.Options.LastEditingIpAddress.displayIndex.Value);
                
                string message = _createRegexRules(csvExportAssessmentOptionsObj.Options.LastEditingIpAddress,
                    nameof(csvExportAssessmentOptionsObj.Options.LastEditingIpAddress));
                
                if (message != null)
                {
                    await Response.WriteAsync( Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, message)));
                    return;
                }
            }
           
           

            #endregion

            foreach (var additionalColumn in csvExportAssessmentOptionsObj.AdditionalCsvColumns)
            {
                displayIndices.Add(new Tuple<int, string, string>(additionalColumn.DisplayIndex, additionalColumn.ArtificialKey, additionalColumn.Name));
                //displayIndicesDict.Add(additionalColumn.ArtificialKey, additionalColumn.DisplayIndex);
            }

            //we need to normalize the indices (from 0, 1, 2 to length) else we cannot directly map to array indices
            var sortedIndices = displayIndices.OrderBy(p => p.Item1).ToList();

            for (int i = 0; i < sortedIndices.Count; i++)
            {
                var tuple = sortedIndices[i];
                 displayIndicesDict[tuple.Item2] = i;
            }


            var release = await _context.ExerciseReleases
                    .Include(p => p.Exercise)
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .ThenInclude(p => p.User)
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .ThenInclude(p => p.Solutions)
                    .ThenInclude(p => p.Assessment)
                    .FirstOrDefaultAsync(p => p.GeneratedCode == releaseCode)
                ;


            if (release == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }


            var pLangs = await _context.PLangs
                .ToListAsync();

            if (pLangs.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no plangs found")));
                return;
            }

            //the user group that manages the exercise
            var targetUserGroup =
                await _context.ExerciseReleases.Where(p => p.Id == release.Id)
                    .Select(p => p.Exercise.UserGroupId)
                    .FirstOrDefaultAsync();

            if (!await base.HasGroupPermission(targetUserGroup,
                permission => permission != null && permission.CanAssessExercises))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            //get all
            var paginationData = new PaginationPostData()
            {
                Page = 1,
                PageSize = -1,
                IsOrderByDesc = null,
                OrderByProp = null,
                SearchText = ""
            };
            var assessmentResults = await this.GetAssessmentResults(release, paginationData);
            var assessments = assessmentResults.Items;


            string fileName = "submissionResults_" + release.GeneratedCode;

            if (!Files.ValidFileName(fileName))
            {
                fileName = "submissionResults";
            }

            fileName += ".csv";


            //http://joshclose.github.io/CsvHelper/writing/
            var builder = new StringBuilder();
            var csvWriter = new CsvWriter(new StringWriter(builder), new CsvConfiguration() {
                Delimiter = csvDelimiter.Trim()
            });

            //additional columns are also inside the sortedIndices
            foreach (var columnName in sortedIndices)
                csvWriter.WriteField(columnName.Item3);

            
            csvWriter.NextRecord();

            csvWriter.Configuration.HasHeaderRecord = true;

            //sort like frontend
            var sortedAssessments = IsOrderByDesc
                ?  assessments
                    .OrderByDescending(sortFunc)
                    .ToList()
                : assessments
                    .OrderBy(sortFunc)
                    .ToList()
                ;


            string[] values = new string[displayIndices.Count]; //here additional cols are already included

            foreach (var assessment in sortedAssessments)
            {
                
                #region known columns
                
                if (csvExportAssessmentOptionsObj.Options.Token.willExport && csvExportAssessmentOptionsObj.Options.Token.displayIndex.HasValue)
                {
                    string val = assessment.Token;
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.Token, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.Token)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.FirstName.willExport  && csvExportAssessmentOptionsObj.Options.FirstName.displayIndex.HasValue)
                {
                    string val = assessment.FirstName;
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.FirstName, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.FirstName)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.LastName.willExport  && csvExportAssessmentOptionsObj.Options.LastName.displayIndex.HasValue)
                {
                    string val = assessment.LastName;
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.LastName, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.LastName)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.ShouldNotCount.willExport  && csvExportAssessmentOptionsObj.Options.ShouldNotCount.displayIndex.HasValue)
                {
                    string val = assessment.ShouldNotCount ? "true" : "false";
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.ShouldNotCount, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.ShouldNotCount)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.UserId.willExport  && csvExportAssessmentOptionsObj.Options.UserId.displayIndex.HasValue)
                {
                    string val = assessment.UserId.ToString();
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.UserId, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.UserId)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.PLangId.willExport  && csvExportAssessmentOptionsObj.Options.PLangId.displayIndex.HasValue) {
                    var _plang = pLangs.FirstOrDefault(p => p.Id == assessment.PLangId);
                    string val = _plang == null
                        ? "null"
                        : _plang.DisplayName;
                    
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.PLangId, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.PLangId)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.ExerciseReleaseId.willExport && csvExportAssessmentOptionsObj.Options.ExerciseReleaseId.displayIndex.HasValue)
                {
                    string val = assessment.ExerciseReleaseId.ToString();
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.ExerciseReleaseId, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.ExerciseReleaseId)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.ParticipationCount.willExport  && csvExportAssessmentOptionsObj.Options.ParticipationCount.displayIndex.HasValue)
                {
                    string val = assessment.ParticipationCount.ToString();
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.ParticipationCount, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.ParticipationCount)]] = val;
                }
                
                if (csvExportAssessmentOptionsObj.Options.SubmitTestPoints.willExport  && csvExportAssessmentOptionsObj.Options.SubmitTestPoints.displayIndex.HasValue)
                {
                    string val = assessment.SubmitTestPoints.HasValue 
                        ? assessment.SubmitTestPoints.Value.ToString() 
                        : "";
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.SubmitTestPoints, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.SubmitTestPoints)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints.willExport  && csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints.displayIndex.HasValue)
                {
                    string val = assessment.MaxSubmitTestPoints.HasValue 
                        ? assessment.MaxSubmitTestPoints.Value.ToString() 
                        : "";
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.NormalTestPoints.willExport  && csvExportAssessmentOptionsObj.Options.NormalTestPoints.displayIndex.HasValue)
                {
                    string val = assessment.NormalTestPoints.HasValue 
                        ? assessment.NormalTestPoints.Value.ToString() 
                        : "";
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.MaxSubmitTestPoints, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.NormalTestPoints)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints.willExport  && csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints.displayIndex.HasValue)
                {
                    string val = assessment.MaxNormalTestPoints.HasValue 
                        ? assessment.MaxNormalTestPoints.Value.ToString() 
                        : "";
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.MaxNormalTestPoints)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.ManualPoints.willExport  && csvExportAssessmentOptionsObj.Options.ManualPoints.displayIndex.HasValue)
                {
                    string val = assessment.ManualPoints.HasValue 
                        ? assessment.ManualPoints.Value.ToString() 
                        : "";
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.ManualPoints, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.ManualPoints)]] = val;
                }
                if (csvExportAssessmentOptionsObj.Options.MaxManualPoints.willExport  && csvExportAssessmentOptionsObj.Options.MaxManualPoints.displayIndex.HasValue)
                {
                    string val = assessment.MaxManualPoints.HasValue 
                        ? assessment.MaxManualPoints.Value.ToString() 
                        : "";
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.MaxManualPoints, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.MaxManualPoints)]] = val;
                }
                
                if (csvExportAssessmentOptionsObj.Options.NoteForOtherTutors.willExport  && csvExportAssessmentOptionsObj.Options.NoteForOtherTutors.displayIndex.HasValue)
                {
                    string val = assessment.NoteForOtherTutors == null ? "" : assessment.NoteForOtherTutors;
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.NoteForOtherTutors, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.NoteForOtherTutors)]] = val;
                }
                
                
                if (csvExportAssessmentOptionsObj.Options.FeedbackForStudent.willExport  && csvExportAssessmentOptionsObj.Options.FeedbackForStudent.displayIndex.HasValue)
                {
                    string val = assessment.FeedbackForStudent == null ? "" : assessment.FeedbackForStudent;
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.FeedbackForStudent, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.FeedbackForStudent)]] = val;
                }
                
                if (csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage.willExport  && csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage.displayIndex.HasValue)
                {
                    string val = assessment.LastAssessmentErrorMessage == null ? "" : assessment.LastAssessmentErrorMessage;
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.LastAssessmentErrorMessage)]] = val;
                }
                
                if (csvExportAssessmentOptionsObj.Options.LastEditingIpAddress.willExport  && csvExportAssessmentOptionsObj.Options.LastEditingIpAddress.displayIndex.HasValue)
                {
                    string val = assessment.LastEditingIpAddress == null ? "" : assessment.LastEditingIpAddress;
                    val = _applyReplaceRule(csvExportAssessmentOptionsObj.Options.LastEditingIpAddress, val);
                    values[displayIndicesDict[nameof(csvExportAssessmentOptionsObj.Options.LastEditingIpAddress)]] = val;
                }
              
                #endregion

                //then insert additional cols
                foreach (var additionalCol in csvExportAssessmentOptionsObj.AdditionalCsvColumns)
                {
                    values[displayIndicesDict[additionalCol.ArtificialKey]] = additionalCol.Value;
                }

                foreach (var item in values)
                {
                     csvWriter.WriteField(item);
                }

                csvWriter.NextRecord();
            }

            csvWriter.Dispose();

            Response.ContentType = "text/comma-separated-values";
            Response.Headers.Add(new KeyValuePair<string, StringValues>("Content-Disposition",
                "attachment; filename=\"" + fileName + "\""));
            await Response.WriteAsync(builder.ToString(), UTF8WithoutBom);
        }

        private string _applyReplaceRule(CsvExportAssessmentOptionTuple tuple, string val)
        {
            bool matched = false;
            foreach (var replaceRule in tuple.ReplaceRulesObj.ReplaceRules)
            {
                if (replaceRule.RegexCompiled.IsMatch(val))
                {
                    val = replaceRule.ReplaceWith;
                    matched = true;
                    break;
                }
            }

            if (matched == false && tuple.ReplaceRulesObj.ElseReplace != null)
            {
                val = tuple.ReplaceRulesObj.ElseReplace;
            }

            return val;
        }

        private string _createRegexRules(CsvExportAssessmentOptionTuple tuple, string columnName)
        {
            foreach (var replaceRule in tuple.ReplaceRulesObj.ReplaceRules)
            {
                try
                {

                    if (replaceRule.Regex == "")
                    {
                        throw new Exception("empty regex"); //we don't allow this
                    }
                    
                    replaceRule.RegexCompiled = new Regex(replaceRule.Regex, _parseRegexFlags(replaceRule.RegexFlags));
                }
                catch (Exception e)
                {
                    return "invalid replace rule (regex) for column: '" + columnName + "', regex: '" + replaceRule.Regex +
                           "', flags: '" + replaceRule.RegexFlags + "'";
                }
                    
            }

            return null;
        }

        private RegexOptions _parseRegexFlags(string flags)
        {
            RegexOptions options = RegexOptions.None;
            if (flags.Contains("i"))
            {
                options |= RegexOptions.IgnoreCase;
            }

            if (flags.Contains("m"))
            {
                options |= RegexOptions.Multiline;
            }

            if (flags.Contains("g"))
            {
                //do nothing
            }

            //make js compatible
            options |= RegexOptions.ECMAScript;

            return options;
        }

        /// <summary>
        /// the file needs at least all columns in <see cref="InternalCsvImportAssessment"/>
        /// except <see cref="InternalCsvImportAssessment.ExerciseReleaseId"/> because that is taken from the release
        /// the first line should be a comment like # [release code]
        /// </summary>
        /// <param name="fileData"></param>
        /// <param name="releaseCode"></param>
        /// <returns></returns>
        [HttpPost("import/csv/{releaseCode}")]
        public async Task ImportCsvAssessments([FromBody] byte[] fileData, string releaseCode)
        {
            if (!await base.IsLoggedIn()) return;

            //the user group that manages the exercise
            var targetUserGroup =
                await _context.ExerciseReleases.Where(p => p.GeneratedCode == releaseCode)
                    .Select(p => p.Exercise.UserGroupId)
                    .FirstOrDefaultAsync();

            if (!await base.HasGroupPermission(targetUserGroup,
                permission => permission != null && permission.CanAssessExercises))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            var release = await _context.ExerciseReleases
                        .Include(p => p.Exercise)
                    .FirstOrDefaultAsync(p => p.GeneratedCode == releaseCode)
                ;


            if (release == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }
            
            if (release.Exercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (release.Exercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "exercise is permanently locked")));
                return;
            }

            string fileContent = Encoding.UTF8.GetString(fileData);

            bool isRightReleaseCode = false;
            int firstLineComment = fileContent.IndexOf('\n');
            if (firstLineComment >= 0) {
                string firstLine = fileContent.Substring(0, firstLineComment); //could include \r from \r\n
                string trimmedFirstLine = firstLine.Substring(1).Trim(); //csv comment is always a char --> length 1

                if (trimmedFirstLine.Length >= release.GeneratedCode.Length) {
                    string csvReleaseCode = trimmedFirstLine.Substring(0, release.GeneratedCode.Length);

                    if (csvReleaseCode == release.GeneratedCode) {
                        isRightReleaseCode = true;
                    }
                }
            }

            if (isRightReleaseCode == false) {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "csv file release code was not equal to the target release (code), maybe you chose the wrong file?")));
                return;
            }

            var solutions = await _context.Solutions
                    .Include(p => p.Assessment)
                    .Where(p => p.ExerciseReleaseId == release.Id)
                    .ToListAsync()
                ;

            var reader = new CsvReader(new StringReader(fileContent));
            reader.Configuration.Delimiter = CsvDelimiterChar;
            reader.Configuration.HasHeaderRecord = true;
            reader.Configuration.Comment = CsvCommentChar;
            reader.Configuration.AllowComments = true;

            //var list = reader.GetRecords<CsvExportImportAssessment>().ToList();

            //not all fields we gave the frontend are used here... (because they are readonly)

            int lineCount = 1;
            var csvImportList = new List<InternalCsvImportAssessment>();

            try
            {
                while (reader.Read())
                {
                    var userIdString = reader.GetField(nameof(CsvExportImportAssessment.UserId));

                    if (int.TryParse(userIdString, out var userId) == false)
                    {
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                    "could not parse user id on line " + lineCount)));
                        return;
                    }

                    var pLangString = reader.GetField(nameof(CsvExportImportAssessment.PLangId));

                    if (int.TryParse(pLangString, out var pLangId) == false)
                    {
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                    "could not parse plang id on line " + lineCount)));
                        return;
                    }

                    var manualPointsString = reader.GetField(nameof(CsvExportImportAssessment.ManualPoints));
                    int? manualPoints = null;

                    if (manualPointsString == null || manualPointsString == "")
                    {
                        //allowed
                    }
                    else
                    {
                        if (int.TryParse(manualPointsString, out var _manualPoints) == false)
                        {
                            await
                                Response.WriteAsync(
                                    Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                        "could not parse manual points on line " + lineCount)));
                            return;
                        }

                        manualPoints = _manualPoints;
                    }

                    var feedBackForStudent = reader.GetField(nameof(CsvExportImportAssessment.FeedbackForStudent));
                    var feedBackForOtherTutors = reader.GetField(nameof(CsvExportImportAssessment.NoteForOtherTutors));

                    csvImportList.Add(new InternalCsvImportAssessment()
                    {
                        UserId = userId,
                        PLangId = pLangId,
                        ManualPoints = manualPoints,
                        FeedbackForStudent = feedBackForStudent,
                        NoteForOtherTutors = feedBackForOtherTutors
                    });
                    lineCount++;
                }
            }
            catch (System.Exception ex)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "could not process line " + lineCount + ", error: " + ex.Message)));
                return;
            }

            foreach (var csvAssessment in csvImportList)
            {
                var solution = solutions.FirstOrDefault(p =>
                    p.UserId == csvAssessment.UserId && p.PLangId == csvAssessment.PLangId);

                if (solution == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "could not find solution for user id:" + csvAssessment.UserId + ", plang id:" +
                                csvAssessment.PLangId + ", release id:" + release.Id)));
                    return;
                }


                if (solution.Assessment == null)
                {
                    //do nothing...
                    Console.WriteLine("skipped assessment for user id:" + csvAssessment.UserId + ", plang id:" +
                                      csvAssessment.PLangId + ", release id:" + release.Id + ", because assessment was null");
                    continue;
                }

                //take values
                solution.Assessment.ManualPoints = csvAssessment.ManualPoints;
                solution.Assessment.FeedbackForStudent = csvAssessment.FeedbackForStudent;
                solution.Assessment.NoteForOtherTutors = csvAssessment.NoteForOtherTutors;
            }

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (System.Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
            return;
        }

        /// <summary>
        /// downloads a single submission as a zip file
        /// </summary>
        /// <param name="releaseCode"></param>
        /// <param name="targetUserId"></param>
        /// <param name="pLangId"></param>
        [HttpGet("download/submission/{releaseCode}/{targetUserId}/{pLangId}")]
        public async Task DownloadSubmissionAsZip(string releaseCode, int targetUserId, int pLangId)
        {
            if (!await base.IsLoggedIn(null, false, false)) return;

            int userId = GetUserId();

            //the user group that manages the exercise
            var targetUserGroup =
                await _context.ExerciseReleases.Where(p => p.GeneratedCode == releaseCode)
                    .Select(p => p.Exercise.UserGroupId)
                    .FirstOrDefaultAsync();

            if (!await base.HasGroupPermission(targetUserGroup,
                permission => permission != null && permission.CanAssessExercises))
            {
                Response.StatusCode = 406; //client expects zip but we send back json
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            //just ensure that the downloading user still exists
            var downloadingUser = await _context.Users.FirstOrDefaultAsync(p => p.Id == userId);
            if (downloadingUser == null)
            {
                Response.StatusCode = 406; 
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no user found")));
                return;
            }
            
            var targetUser = await _context.Users.FirstOrDefaultAsync(p => p.Id == targetUserId);

            if (targetUser == null)
            {
                Response.StatusCode = 406;
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no user found")));
                return;
            }

            //get all automatic assessments for this release
            var solution = await _context.Solutions
                    .Include(p => p.SolutionFiles)
                    .FirstOrDefaultAsync(p => p.UserId == targetUserId
                                              &&
                                              p.ExerciseReleaseWithUserAsParticipation.ExerciseRelease.GeneratedCode ==
                                              releaseCode
                                              && p.PLangId == pLangId)
                ;


            if (solution == null)
            {
                Response.StatusCode = 406; 
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "solution not found")));
                return;
            }

            #region make sure the readonly files are up-to-date

            var release = await _context.ExerciseReleases
                    .Include(p => p.Exercise.CodeTemplates)
                    .ThenInclude(p => p.TemplateFiles)
                    .FirstOrDefaultAsync(p => p.GeneratedCode == releaseCode)
                ;
            
            if (release == null)
            {
                Response.StatusCode = 406;
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            var codeTemplate = release.Exercise.CodeTemplates
                    .FirstOrDefault(p => p.PLangId == solution.PLangId)
                ;
            
            if (codeTemplate == null)
            {
                Response.StatusCode = 406;
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "codeTemplate not found")));
                return;
            }

            var errorMsg = UserSolutionHelper.ReplaceReadonlyFiles(solution, codeTemplate, false, true);

            if (errorMsg != null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, errorMsg)));
                return;
            }
            
            #endregion

            

            var memStream = new MemoryStream();

            var zipArchive = new ZipArchive(memStream, ZipArchiveMode.Create, true, UTF8WithoutBom);

            string dirName = "submission";

            //some user might have 2 first or 2 last names --> replace with _
            string betterName =
                Files.ReplaceUmlauts(targetUser.LastName.Replace(" ", "_") + "_" +
                                     targetUser.FirstName.Replace(" ", "_"));

            if (Files.ValidFileName(betterName))
            {
                dirName = dirName + "_" + betterName;
            }


            int count = 0;
            foreach (var solutionFile in solution.SolutionFiles)
            {
                try
                {
                    var fileName = solutionFile.FileNameWithExtension;

                    if (!Files.ValidFileName(solutionFile.FileNameWithExtension))
                    {
                        fileName = "invalidFileName" + count++;
                    }

                    var entry = zipArchive.CreateEntry(dirName + "/" + fileName, CompressionLevel.Fastest);

                    using (var streamWriter = new StreamWriter(entry.Open(), UTF8WithoutBom))
                    {
                        streamWriter.Write(solutionFile.Content);
                    }
                }
                catch (Exception ex)
                {
                    if (AppConfiguration.IsDebugMode)
                    {
                        Response.StatusCode = 406;
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                    "error when zipping file: " + solutionFile.FileNameWithExtension + ", message: " +
                                    ex.Message)));
                    }
                    else
                    {
                        Response.StatusCode = 406;
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                    "at least 1 file could not be zipped")));
                    }

                    return;
                }
            }

            zipArchive.Dispose();
            //Response.ContentType = "application/zip";
            //Response.Headers.Add(new KeyValuePair<string, StringValues>("Content-Disposition","attachment; filename=" + dirName + ".zip"));


            //var arr = memStream.ToArray();

            //Response.Body.Write(arr, 0, arr.Length);


            //await Response.Body.FlushAsync();

            var t = base.File(memStream.ToArray(), "application/zip", dirName + ".zip");

            await t.ExecuteResultAsync(this.ControllerContext);
        }

        /// <summary>
        /// returns all submissions as a zip file (download)
        ///
        /// include the csv file for the tutors that can then be filled and uploaded again
        /// for every plang we create a separate directory (better sorted)
        /// we also include in every plang direcotry a default submission (with all template files) to be able to compare submissions against the code templates
        ///
        /// include also the file name for the submission in the csv file e.g in case we needed to change the filename (ä,ö,ü, ... bad for cross platform filenames)
        /// </summary>
        /// <param name="releaseCode"></param>
        [HttpGet("download/submission/all/{releaseCode}")]
        public async Task DownloadAllSubmissionsAsZip(string releaseCode)
        {
            if (!await base.IsLoggedIn(null, false, false)) return;

            int userId = GetUserId();


            //the user group that manages the exercise
            var targetUserGroup =
                await _context.ExerciseReleases.Where(p => p.GeneratedCode == releaseCode)
                    .Select(p => p.Exercise.UserGroupId)
                    .FirstOrDefaultAsync();

            if (!await base.HasGroupPermission(targetUserGroup,
                permission => permission != null && permission.CanAssessExercises))
            {
                Response.StatusCode = 406; //client expects zip but we send json back
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            var user = await _context.Users.FirstOrDefaultAsync(p => p.Id == userId);

            if (user == null)
            {
                Response.StatusCode = 406; //client expects zip but we send back json
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no user found")));
                return;
            }

            //get all automatic assessments for this release
            //TODO maybe performance ...
            var solutions = await _context.Solutions
                    .Include(p => p.Assessment)
                    .Include(p => p.SolutionFiles)
                    .Include(p => p.PLang)
                    .Include(p => p.ExerciseReleaseWithUserAsParticipation)
                    .ThenInclude(p => p.User)
                    .Where(
                        p =>
                            //p.UserId == userId  && when we our solution...
                            p.ExerciseReleaseWithUserAsParticipation.ExerciseRelease.GeneratedCode ==
                            releaseCode)
                    .OrderBy(p => p.ExerciseReleaseWithUserAsParticipation.User.LastName)
                    .ToListAsync()
                ;

            var release = await _context.ExerciseReleases
                .Include(p => p.Exercise.CodeTemplates)
                .ThenInclude(p => p.TemplateFiles)
                .FirstOrDefaultAsync(p => p.GeneratedCode == releaseCode)
                ;

            if (release == null)
            {
                Response.StatusCode = 406;
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            var participationCounts = await _context.ExerciseReleaseWithUserAsParticipations
                    .Where(p => p.ExerciseRelease.ExerciseId == release.ExerciseId)
                    .GroupBy(p => p.UserId)
                    .Select(p => new
                    {
                        UserId = p.Key,
                        ParticipationCount = p.Count()
                    })
                    .ToDictionaryAsync(p => p.UserId, p => p.ParticipationCount)
                ;

            var memStream = new MemoryStream();

            var zipArchive = new ZipArchive(memStream, ZipArchiveMode.Create, true, UTF8WithoutBom);

            string dirName = "submissions";

            if (Files.ValidFileName(dirName + "_" + release.GeneratedCode))
            {
                dirName = dirName + "_" + release.GeneratedCode;
            }


            var csvExportAssessments = new List<CsvExportImportAssessment>();
            //the plangs we have submissions for
            var usedPLangIds = new HashSet<int>();
            var usedPLangs = new List<PLang>();
            for (int i = 0; i < solutions.Count; i++)
            {
                Solution solution = solutions[i];

                if (participationCounts.TryGetValue(solution.UserId, out int participationCount) == false) {
                    Response.StatusCode = 406; 
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "could not find participation count for user id: " + solution.UserId + " for release id: " + release.Id + ", plangId: " + solution.PLangId)));
                }
                
                #region make sure the readonly files are up-to-date
                
                var codeTemplate = release.Exercise.CodeTemplates
                        .FirstOrDefault(p => p.PLangId == solution.PLangId)
                    ;
            
                if (codeTemplate == null)
                {
                    Response.StatusCode = 406;
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "could not find codeTemplate for user id: " + solution.UserId + " for release id: " + release.Id + ", plangId: " + solution.PLangId)));
                    return;
                }

               var errorMsg = UserSolutionHelper.ReplaceReadonlyFiles(solution, codeTemplate, false, true);

                if (errorMsg != null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.ServerError, errorMsg)));
                    return;
                }
                
                #endregion
                
                
                csvExportAssessments.Add(new CsvExportImportAssessment()
                {
                    UserId = solution.UserId,
                    PLangId = solution.PLangId,
                    PLangName = solution.PLang.DisplayName,
                    NoteForOtherTutors = solution.Assessment?.NoteForOtherTutors,
                    FeedbackForStudent = solution.Assessment?.FeedbackForStudent,
                    ShouldNotCount = solution.ExerciseReleaseWithUserAsParticipation.ShouldNotCount,
                    ManualPoints = solution.Assessment?.ManualPoints,
                    Token = solution.ExerciseReleaseWithUserAsParticipation.User.Token,
                    SubmissionDirName = "will be set",
                    LastName = solution.ExerciseReleaseWithUserAsParticipation.User.LastName,
                    FirstName = solution.ExerciseReleaseWithUserAsParticipation.User.FirstName,
                    ParticipationCount = participationCount,
                    NormalTestPoints = solution.Assessment?.NormalTestPoints,
                    SubmitTestPoints = solution.Assessment?.SubmitTestPoints,
                    LastAssessmentErrorMessage = solution.Assessment?.LastAssessmentErrorMessage,
                    MaxNormalTestPoints = solution.Assessment?.MaxNormalTestPoints,
                    MaxSubmitTestPoints = solution.Assessment?.MaxSubmitTestPoints,
                    MaxManualPoints = release.MaxManualPoint,
                });


                string plangDirName = solution.PLang.InternalName;

                if (usedPLangIds.Add(solution.PLang.Id)) {
                    usedPLangs.Add(solution.PLang);
                }

                string submissionDirName = "submission-";

                //some user might have 2 first or 2 last names --> replace with _
                string betterName =
                    Files.ReplaceUmlauts(
                        solution.ExerciseReleaseWithUserAsParticipation.User.LastName.Replace(" ", "_") + "-" +
                        solution.ExerciseReleaseWithUserAsParticipation.User.FirstName.Replace(" ", "_"));


                //add plangDirName to submissionDirName else we not get the entry in the csv file
                if (Files.ValidFileName(betterName))
                {
                    //include user id here too, we could have to users with the same name...
                    submissionDirName = plangDirName + "/" + submissionDirName + betterName + "-" + solution.ExerciseReleaseWithUserAsParticipation.User.Id;
                }
                else
                {
                    submissionDirName = plangDirName + "/" + submissionDirName + solution.ExerciseReleaseWithUserAsParticipation.User.Id;
                }

                csvExportAssessments[csvExportAssessments.Count - 1].SubmissionDirName = submissionDirName;


                int count = 0;
                foreach (var solutionFile in solution.SolutionFiles)
                {
                    try
                    {
                        var fileName = solutionFile.FileNameWithExtension;

                        if (!Files.ValidFileName(solutionFile.FileNameWithExtension))
                        {
                            fileName = "invalidFileName" + count++;
                        }

                        var entry = zipArchive.CreateEntry(dirName + "/" + submissionDirName + "/" + fileName,
                            CompressionLevel.Fastest);
                        using (var streamWriter = new StreamWriter(entry.Open(), UTF8WithoutBom))
                        {
                            streamWriter.Write(solutionFile.Content);
                        }
                    }
                    catch (Exception ex)
                    {
                        if (AppConfiguration.IsDebugMode)
                        {
                            Response.StatusCode = 406;
                            await
                                Response.WriteAsync(
                                    Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                        "error when zipping file: " + solutionFile.FileNameWithExtension +
                                        ", message: " +
                                        ex.Message)));
                        }
                        else
                        {
                            Response.StatusCode = 406;
                            await
                                Response.WriteAsync(
                                    Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                        "at least 1 file could not be zipped")));
                        }

                        return;
                    }
                }
            }

            //also add the code template for all plangs we encountered in the submissions

            #region add template default submissions

            foreach (var plang in usedPLangs)
            {
                int count = 0;
                var codeTemplate = release.Exercise.CodeTemplates
                        .FirstOrDefault(p => p.PLangId == plang.Id)
                    ;
            
                if (codeTemplate == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "could not find codeTemplate for default submission, plangId: " + plang.Id)));
                    return;
                }

                foreach (var templateFile in codeTemplate.TemplateFiles)
                {
                    try
                    {
                        var fileName = templateFile.FileNameWithExtension;

                        if (!Files.ValidFileName(templateFile.FileNameWithExtension))
                        {
                            fileName = "invalidFileName" + count++;
                        }

                        var entry = zipArchive.CreateEntry(dirName + "/" + plang.InternalName + "/" + DefaultSubmissionWithCodeTemplatesDirName + "/" + fileName,
                            CompressionLevel.Fastest);
                        using (var streamWriter = new StreamWriter(entry.Open(), UTF8WithoutBom))
                        {
                            streamWriter.Write(templateFile.Content);
                        }
                    }
                    catch (Exception ex)
                    {
                        if (AppConfiguration.IsDebugMode)
                        {
                            await
                                Response.WriteAsync(
                                    Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                        "error when zipping file: " + templateFile.FileNameWithExtension +
                                        ", message: " +
                                        ex.Message)));
                        }
                        else
                        {
                            await
                                Response.WriteAsync(
                                    Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                                        "at least 1 file could not be zipped")));
                        }

                        return;
                    }
                }
            }
            #endregion

            //we use csv helper for dotnet core ...https://joshclose.github.io/CsvHelper/reading#getting-all-records
            
            var assessmentCsvFile = zipArchive.CreateEntry(dirName + "/assessments_" + release.GeneratedCode + ".csv",
                CompressionLevel.Fastest);

            //the csv file with all data (like the table in the frontend)
            //this can be filled by thetutor and uploaded again
            //the first line is a comment # [release code] to make sure we upload the right file
            using (var streamWriter = new StreamWriter(assessmentCsvFile.Open(), UTF8WithoutBom))
            {
               var writer = new CsvWriter(streamWriter);
                writer.Configuration.Delimiter = CsvDelimiterChar;
                writer.Configuration.Comment = CsvCommentChar;
                writer.Configuration.AllowComments = true;
                writer.WriteField(CsvCommentChar + release.GeneratedCode);
                //writer.WriteComment(); //not found... outdated lib version?
                writer.NextRecord();
                writer.WriteHeader<CsvExportImportAssessment>();
                writer.WriteRecords(csvExportAssessments);  
            }

            zipArchive.Dispose();
            var t = base.File(memStream.ToArray(), "application/zip", dirName + ".zip");

            await t.ExecuteResultAsync(this.ControllerContext);
        }

        /// <summary>
        /// returns the assessment results for the given release 
        /// 
        /// </summary>
        /// <param name="release">needs all ExerciseReleaseWithUserAsParticipations attached</param>
        /// <returns>the assessments</returns>
        private async Task<PaginatedData<AssessmentFullBase>> GetAssessmentResults(ExerciseRelease release,
            PaginationPostData pagination)
        {
            //all participations where is the right exercise
            //group by userid to get the amount of participations
            var participationCounts = await _context.ExerciseReleaseWithUserAsParticipations
                    .Where(p => p.ExerciseRelease.ExerciseId == release.ExerciseId)
                    .GroupBy(p => p.UserId)
                    .Select(p => new
                    {
                        UserId = p.Key,
                        ParticipationCount = p.Count()
                    })
                    .ToDictionaryAsync(p => p.UserId, p => p.ParticipationCount)
                ;


            var assessmentsForFrontend = new List<AssessmentFullBase>();

            foreach (var participation in release.ExerciseReleaseWithUserAsParticipations)
            {
                //user could submit solutions for more than 1 plang
                foreach (var solution in participation.Solutions)
                {
                    int participationCount = -1;

                    //the check should not be necessary...
                    if (participationCounts.TryGetValue(participation.UserId, out participationCount))
                    {
                    }
                    else
                    {
                        Console.WriteLine("WARNING (submission controller): tried to get participation count for: " +
                                          participation.UserId +
                                          " for exercise: " + participation.ExerciseRelease.ExerciseId +
                                          " for release: " +
                                          participation.ExerciseReleaseId);
                    }

                    var assessment = new AssessmentFullBase()
                    {
                        Token = participation.User.Token,
                        FirstName = participation.User.FirstName,
                        LastName = participation.User.LastName,
                        UserId = participation.User.Id,
                        PLangId = solution.PLangId,
                        ExerciseReleaseId = participation.ExerciseReleaseId,
                        ManualPoints = solution.Assessment?.ManualPoints,
                        MaxNormalTestPoints = solution.Assessment?.MaxNormalTestPoints,
                        MaxSubmitTestPoints = solution.Assessment?.MaxSubmitTestPoints,
                        NormalTestPoints = solution.Assessment?.NormalTestPoints,
                        SubmitTestPoints = solution.Assessment?.SubmitTestPoints,
                        NoteForOtherTutors = solution.Assessment?.NoteForOtherTutors,
                        FeedbackForStudent = solution.Assessment?.FeedbackForStudent,
                        ShouldNotCount = participation.ShouldNotCount,
                        ParticipationCount = participationCount,
                        LastAssessmentErrorMessage = solution.Assessment?.LastAssessmentErrorMessage,
                        LastEditingIpAddress = solution.LastEditingIpAddress,
                        MaxManualPoints = release.MaxManualPoint,
                    };

                    assessmentsForFrontend.Add(assessment);
                }
            }

            var frontendData = assessmentsForFrontend.AsQueryable();
//            var count = frontendData.Count();

            //TODO this is no real pagination... we calculate all data and then do pagination = no performance improvement
            //but currently i have no idea how we would do proper pagination here...

            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                frontendData = frontendData
                    .Where(p => (p.FirstName != null && p.FirstName.ToLower().Contains(text))
                                || (p.LastName != null && p.LastName.ToLower().Contains(text))
                                || (p.NoteForOtherTutors != null && p.NoteForOtherTutors.ToLower().Contains(text))
                                || (p.FeedbackForStudent != null && p.FeedbackForStudent.ToLower().Contains(text))
                    );
            }

//            var count2 = frontendData.Count();

            string orderByProp = "";
            Expression<Func<AssessmentFullBase, object>> orderByFunc = null;

            //not all props are sortable via the frontend
            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "FirstName".ToLower())
                {
                    orderByFunc = p => p.FirstName;
                    orderByProp = "FirstName".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "LastName".ToLower())
                {
                    orderByFunc = p => p.LastName;
                    orderByProp = "LastName".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "PLangId".ToLower())
                {
                    orderByFunc = p => p.PLangId;
                    orderByProp = "IsOnlyVisibleToMe".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "ShouldNotCount".ToLower())
                {
                    orderByFunc = p => p.ShouldNotCount;
                    orderByProp = "ShouldNotCount".ToLower();
                }

                else if (pagination.OrderByProp.ToLower() == "ParticipationCount".ToLower())
                {
                    orderByFunc = p => p.ParticipationCount;
                    orderByProp = "ParticipationCount".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "ManualPoints".ToLower())
                {
                    orderByFunc = p => p.ManualPoints;
                    orderByProp = "ManualPoints".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "MaxManualPoints".ToLower())
                {
                    orderByFunc = p => p.MaxManualPoints;
                    orderByProp = "MaxManualPoints".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "NormalTestPoints".ToLower())
                {
                    orderByFunc = p => p.NormalTestPoints;
                    orderByProp = "NormalTestPoints".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "SubmitTestPoints".ToLower())
                {
                    orderByFunc = p => p.SubmitTestPoints;
                    orderByProp = "SubmitTestPoints".ToLower();
                }

                //not needed
//                else if (pagination.OrderByProp.ToLower() == "FeedbackForStudent".ToLower())
//                {
//                    orderByFunc = p => p.FeedbackForStudent;
//                    orderByProp = "FeedbackForStudent".ToLower();
//                }
//                else if (pagination.OrderByProp.ToLower() == "NoteForOtherTutors".ToLower())
//                {
//                    orderByFunc = p => p.NoteForOtherTutors;
//                    orderByProp = "NoteForOtherTutors".ToLower();
//                }
                //& some others e.g. last error message & last editing ip...
            }

            return PaginationHelper.GetAsPagination(frontendData,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );
        }

        /// <summary>
        /// returns the pre results for the given release 
        ///     (the results of the normal test from the user solution are taken, submissions tests are ignored) 
        /// </summary>
        /// <param name="release">needs all ExerciseReleaseWithUserAsParticipations attached</param>
        /// <param name="normalTests"></param>
        /// <returns>the assessments</returns>
        private async Task<PaginatedData<AssessmentFullBase>> GetPreAssessmentResults(ExerciseRelease release,
            List<Test> normalTests, PaginationPostData pagination)
        {
            //all participations where is the right exercise
            //group by userid to get the amount of participations
            var participationCounts = await _context.ExerciseReleaseWithUserAsParticipations
                    .Where(p => p.ExerciseRelease.ExerciseId == release.ExerciseId)
                    .GroupBy(p => p.UserId)
                    .Select(p => new
                    {
                        UserId = p.Key,
                        ParticipationCount = p.Count()
                    })
                    .ToDictionaryAsync(p => p.UserId, p => p.ParticipationCount)
                ;

            int summedNormalTestPoints = normalTests.Sum(p => p.Weight);

            var assessmentsForFrontend = new List<AssessmentFullBase>();

            foreach (var participation in release.ExerciseReleaseWithUserAsParticipations)
            {
                //user could submit solutions for more than 1 plang
                foreach (var solution in participation.Solutions)
                {
                    int participationCount = -1;

                    //the check should not be necessary...
                    if (participationCounts.TryGetValue(participation.UserId, out participationCount))
                    {
                    }
                    else
                    {
                        Console.WriteLine("WARNING (submission controller): tried to get participation count for: " +
                                          participation.UserId +
                                          " for exercise: " + participation.ExerciseRelease.ExerciseId +
                                          " for release: " +
                                          participation.ExerciseReleaseId);
                    }

                    if (solution == null)
                    {
                        Console.WriteLine("WARNING (submission controller): solution not found for: userId" +
                                          participation.UserId +
                                          " for exerciseId: " + participation.ExerciseRelease.ExerciseId +
                                          " for releaseId: " +
                                          participation.ExerciseReleaseId);

                        continue;
                    }

                    int? normalPoints = null;

                    foreach (var normalTest in normalTests)
                    {
                        var testResult = solution.TestResults.Find(p => p.TestId == normalTest.Id);

                        if (testResult == null || testResult.Passed == null)
                        {
                            //not yet run
                        }
                        else
                        {
                            if (normalPoints == null) normalPoints = 0;

                            if (testResult.Passed.Value)
                            {
                                normalPoints += normalTest.Weight;
                            }
                        }
                    }

                    var assessment = new AssessmentFullBase()
                    {
                        Token = participation.User.Token,
                        FirstName = participation.User.FirstName,
                        LastName = participation.User.LastName,
                        UserId = participation.User.Id,
                        PLangId = solution.PLangId,
                        ExerciseReleaseId = participation.ExerciseReleaseId,
                        ManualPoints = solution.Assessment?.ManualPoints,
                        MaxNormalTestPoints = summedNormalTestPoints,
                        MaxSubmitTestPoints = null,
                        NormalTestPoints = normalPoints,
                        SubmitTestPoints = null,
                        NoteForOtherTutors = solution.Assessment?.NoteForOtherTutors,
                        FeedbackForStudent = solution.Assessment?.FeedbackForStudent,
                        ShouldNotCount = participation.ShouldNotCount,
                        ParticipationCount = participationCount,
                        LastAssessmentErrorMessage = solution.Assessment?.LastAssessmentErrorMessage,
                        LastEditingIpAddress = solution.LastEditingIpAddress,
                        MaxManualPoints = release.MaxManualPoint
                    };

                    assessmentsForFrontend.Add(assessment);
                }
            }

            var frontendData = assessmentsForFrontend.AsQueryable();

            //TODO this is no real pagination... we calculate all data and then do pagination = no performance improvement
            //but currently i have no idea how we would do proper pagination here...

            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                frontendData = frontendData
                    .Where(p => (p.FirstName != null && p.FirstName.ToLower().Contains(text))
                                || (p.LastName != null && p.LastName.ToLower().Contains(text))
                                || (p.NoteForOtherTutors != null && p.NoteForOtherTutors.ToLower().Contains(text))
                                || (p.FeedbackForStudent != null && p.FeedbackForStudent.ToLower().Contains(text))
                    );
            }

            string orderByProp = "";
            Expression<Func<AssessmentFullBase, object>> orderByFunc = null;

            //not all props are sortable via the frontend
            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "FirstName".ToLower())
                {
                    orderByFunc = p => p.FirstName;
                    orderByProp = "FirstName".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "LastName".ToLower())
                {
                    orderByFunc = p => p.LastName;
                    orderByProp = "LastName".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "PLangId".ToLower())
                {
                    orderByFunc = p => p.PLangId;
                    orderByProp = "IsOnlyVisibleToMe".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "ShouldNotCount".ToLower())
                {
                    orderByFunc = p => p.ShouldNotCount;
                    orderByProp = "ShouldNotCount".ToLower();
                }

                else if (pagination.OrderByProp.ToLower() == "ParticipationCount".ToLower())
                {
                    orderByFunc = p => p.ParticipationCount;
                    orderByProp = "ParticipationCount".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "ManualPoints".ToLower())
                {
                    orderByFunc = p => p.ManualPoints;
                    orderByProp = "ManualPoints".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "MaxManualPoints".ToLower())
                {
                    orderByFunc = p => p.MaxManualPoints;
                    orderByProp = "MaxManualPoints".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "NormalTestPoints".ToLower())
                {
                    orderByFunc = p => p.NormalTestPoints;
                    orderByProp = "NormalTestPoints".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "SubmitTestPoints".ToLower())
                {
                    orderByFunc = p => p.SubmitTestPoints;
                    orderByProp = "SubmitTestPoints".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "FeedbackForStudent".ToLower())
                {
                    orderByFunc = p => p.FeedbackForStudent;
                    orderByProp = "FeedbackForStudent".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "NoteForOtherTutors".ToLower())
                {
                    orderByFunc = p => p.NoteForOtherTutors;
                    orderByProp = "NoteForOtherTutors".ToLower();
                }
            }

            return PaginationHelper.GetAsPagination(frontendData,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );
        }

        //will only work if all release assessments have already finished
        [HttpPost("rerun/{releaseCode}")]
        public async Task RerunAutomaticAssessment(string releaseCode)
        {
            if (!await base.IsLoggedIn()) return;

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }
            
            //get all automatic assessments for this release

            var release = await _context.ExerciseReleases
                    .Include(p => p.Exercise)
                    .FirstOrDefaultAsync(p => p.GeneratedCode == releaseCode)
                ;

            if (release == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }

            if (release.Exercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (release.Exercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "exercise is permanently locked")));
                return;
            }

            //the user group that manages the exercise
            var targetUserGroup =
                await _context.ExerciseReleases.Where(p => p.Id == release.Id)
                    .Select(p => p.Exercise.UserGroupId)
                    .FirstOrDefaultAsync();

            if (!await base.HasGroupPermission(targetUserGroup,
                permission => permission != null && permission.CanAssessExercises))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            var testingController = new TestingController(_context);

            var success =
                await SubmissionAssessmentScheduler.RunForRelease(release, _context, testingController, false, true, systemSettings);

            if (!success)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<bool>(ResponseCode.Ok,
                            "some error occured during the automatic assessment or release is still active or the automatic release has not finished yet", false)));
                return;
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(
                        new BasicResponseWithData<bool>(ResponseCode.Ok, "automatic assessment finished", true)));
        }


        //reruns a single assessment (for a single user) only if the all release assessments have already finished
        [HttpPost("rereunsingle/{releaseCode}/{userId}/{pLangId}")]
        public async Task RerunAutomaticAssessmentForSolution(string releaseCode, int userId, int pLangId)
        {
            if (!await base.IsLoggedIn()) return;

            var systemSettings = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "system settings not found")));
                return;
            }
            
            //get all automatic assessments for this release

            var release = await _context.ExerciseReleases
                    .Include(p => p.Exercise)
                    .FirstOrDefaultAsync(p => p.GeneratedCode == releaseCode)
                ;

            if (release == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "release not found")));
                return;
            }
            
            if (release.Exercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            if (release.Exercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "exercise is permanently locked")));
                return;
            }


            //the user group that manages the exercise
            var targetUserGroup =
                await _context.ExerciseReleases.Where(p => p.Id == release.Id)
                    .Select(p => p.Exercise.UserGroupId)
                    .FirstOrDefaultAsync();

            if (!await base.HasGroupPermission(targetUserGroup,
                permission => permission != null && permission.CanAssessExercises))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            //validate userid and plangid...

            var user = await _context.Users.FirstOrDefaultAsync(p => p.Id == userId);

            if (user == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }

            var plang = await _context.PLangs.FirstOrDefaultAsync(p => p.Id == pLangId);

            if (plang == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "plang not found")));
                return;
            }

            var testingController = new TestingController(_context);
            var solutionPk = new SolutionPrimaryKey()
            {
                PLangId = plang.Id,
                UserId = user.Id,
                ExerciseReleaseId = release.Id,
            };

            var success = await SubmissionAssessmentScheduler.RunSingleAutomaticAssessment(release, solutionPk,
                _context,
                testingController, systemSettings);

            if (success == false)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "some error occured during the automatic assessment or release is still active or the automatic release has not finished yet")));
                return;
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<bool>(ResponseCode.Ok, "automatic assessment finished",
                        success)));
        }
    }


    /// <summary>
    /// an assessment for a submission (user code/program)
    /// we only send this to the frontend but never receive this class ... maybe change to fromBackend?
    /// </summary>
    public class AssessmentFullBase
    {
        /// <summary>
        /// the token from the user
        /// </summary>
        public string Token { get; set; }
        
        /// <summary>
        /// the first name of the connected user
        /// </summary>
        public string FirstName { get; set; }

        /// <summary>
        /// the last name of the connected user
        /// </summary>
        public string LastName { get; set; }

        /// <summary>
        /// this is taken from the release participation so it counts for all p langs
        /// true: the student wants that his/her solution is not counted (for whatever), false: not
        /// can only be set once by the student
        /// </summary>
        public bool ShouldNotCount { get; set; }


        /// <summary>
        /// the user id (in case the tutor wants to see the solution NOT SUPPORTED YET to export as csv)
        /// 
        /// PK
        /// </summary>
        public int UserId { get; set; }

        /// <summary>
        /// the used plang
        /// PK
        /// </summary>
        public int PLangId { get; set; }

        /// <summary>
        /// the connected exercise release id
        /// PK
        /// </summary>
        public int ExerciseReleaseId { get; set; }


        /// <summary>
        /// a user can solve an exercise multiple times (connected through release.exercise)
        /// so this gives us the times the user participated on the exercise (through different releases)
        /// 
        /// IF a user solved the exercise with different p langs then this does not count separate because
        /// all restrictions are stored at the participation (e.g. time) and thus the p lang does not matter!
        /// 
        /// CAN be -1 on error
        /// 
        /// sometimes we want to know if a user solved an exercise multiple times
        /// e.g. when he/she is only allowed to solve it one time (e.g. an exam)
        /// </summary>
        public int ParticipationCount { get; set; }

        /// <summary>
        /// the summed test points from the submit tests
        /// </summary>
        public int? SubmitTestPoints { get; set; }

        /// <summary>
        /// the summed max reachable submit test points 
        /// </summary>
        public int? MaxSubmitTestPoints { get; set; }

        /// <summary>
        /// the summed points from the normal test
        /// </summary>
        public int? NormalTestPoints { get; set; }

        /// <summary>
        /// the summed max reachable normal test points 
        /// </summary>
        public int? MaxNormalTestPoints { get; set; }

        /// <summary>
        /// optional the manual given poitns (by the tutor)
        /// </summary>
        public int? ManualPoints { get; set; }
        
        /// <summary>
        /// the max manual points (just as a note), manual points can be more than this value
        /// </summary>
        public int? MaxManualPoints { get; set; }

        /// <summary>
        /// the note from the tutor
        /// </summary>
        public string NoteForOtherTutors { get; set; }

        /// <summary>
        /// the note for the student why he/she gets the manual points (feedback)
        /// </summary>
        public string FeedbackForStudent { get; set; }

        /// <summary>
        /// null or empty string for no error OR
        /// the error message from the test server or some other source
        /// </summary>
        public string LastAssessmentErrorMessage { get; set; }

        /// <summary>
        /// the ip from which the solution was last modified (saved/resetted/initial created)
        /// or null if no captured (e.g. before we introduced this feature)
        /// </summary>
        public string LastEditingIpAddress { get; set; }
    }

    public class CsvExportAssessmentOptionsObj
    {
        public CsvExportAssessmentOptions Options { get; set; }
        /// <summary>
        /// because c sharp cannot handle the frontend mixing...
        /// we split the frontend obj into two props
        /// these are the additional columns which should be inserted into the CsvExportAssessmentOptions
        /// depending on the displayIndex so we need to merge them back together somehow
        /// </summary>
        public AdditionalCsvColumn[] AdditionalCsvColumns { get; set; }
    }

    public class AdditionalCsvColumn
    {
         /// <summary>
        /// an artificial key so we can distinc the columns
        /// </summary>
        public string ArtificialKey { get; set; }
        public string Name { get; set; }
        public string Value { get; set; }

        public int DisplayIndex { get; set; }
    }
    
    public class CsvExportAssessmentOptions {
        
        /// <summary>
        /// the user token
        /// </summary>
        public CsvExportAssessmentOptionTuple Token { get; set; }
        
        /// <summary>
        /// the first name of the connected user
        /// </summary>
        public CsvExportAssessmentOptionTuple FirstName { get; set; }

        /// <summary>
        /// the last name of the connected user
        /// </summary>
        public CsvExportAssessmentOptionTuple LastName { get; set; }

        /// <summary>
        /// this is taken from the release participation so it counts for all p langs
        /// true: the student wants that his/her solution is not counted (for whatever), false: not
        /// can only be set once by the student
        /// </summary>
        public CsvExportAssessmentOptionTuple ShouldNotCount { get; set; }

        /// <summary>
        /// the user id (in case the tutor wants to see the solution NOT SUPPORTED YET)
        /// 
        /// PK
        /// </summary>
        public CsvExportAssessmentOptionTuple UserId { get; set; }

        /// <summary>
        /// the used plang
        /// PK
        /// this will become the plang display name
        /// </summary>
        public CsvExportAssessmentOptionTuple PLangId { get; set; }

        /// <summary>
        /// the connected exercise release id
        /// PK
        /// </summary>
        public CsvExportAssessmentOptionTuple ExerciseReleaseId { get; set; }


        /// <summary>
        /// a user can solve an exercise multiple times (connected through release.exercise)
        /// so this gives us the times the user participated on the exercise (through different releases)
        /// 
        /// IF a user solved the exercise with different p langs then this does not count separate because
        /// all restrictions are stored at the participation (e.g. time) and thus the p lang does not matter!
        /// 
        /// CAN be -1 on error
        /// 
        /// sometimes we want to know if a user solved an exercise multiple times
        /// e.g. when he/she is only allowed to solve it one time (e.g. an exam)
        /// </summary>
        public CsvExportAssessmentOptionTuple ParticipationCount { get; set; }

        /// <summary>
        /// the summed test points from the submit tests
        /// </summary>
        public CsvExportAssessmentOptionTuple SubmitTestPoints { get; set; }

        /// <summary>
        /// the summed max reachable submit test points 
        /// </summary>
        public CsvExportAssessmentOptionTuple MaxSubmitTestPoints { get; set; }

        /// <summary>
        /// the summed points from the normal test
        /// </summary>
        public CsvExportAssessmentOptionTuple NormalTestPoints { get; set; }

        /// <summary>
        /// the summed max reachable normal test points 
        /// </summary>
        public CsvExportAssessmentOptionTuple MaxNormalTestPoints { get; set; }

        /// <summary>
        /// optional the manual given poitns (by the tutor)
        /// </summary>
        public CsvExportAssessmentOptionTuple ManualPoints { get; set; }
        
        /// <summary>
        /// a hint for the max manual points, manual points can be greater than this value
        /// </summary>
        public CsvExportAssessmentOptionTuple MaxManualPoints { get; set; }
        

        /// <summary>
        /// the note from the tutor
        /// </summary>
        public CsvExportAssessmentOptionTuple NoteForOtherTutors { get; set; }

        /// <summary>
        /// the note for the student why he/she gets the manual points (feedback)
        /// </summary>
        public CsvExportAssessmentOptionTuple FeedbackForStudent { get; set; }

        /// <summary>
        /// null or empty string for no error OR
        /// the error message from the test server or some other source
        /// </summary>
        public CsvExportAssessmentOptionTuple LastAssessmentErrorMessage { get; set; }
        
        /// <summary>
        /// the last ip address when a user created or changed the solution
        /// can be null if not captured (before we introduced this prop)
        /// </summary>
        public CsvExportAssessmentOptionTuple LastEditingIpAddress { get; set; }
    }

    public class CsvExportAssessmentOptionTuple {
        /// <summary>
        /// the display index might not start with 0
        /// we only know the order from the indices... needs to be normalized to be used (0 to number of tuples)
        /// </summary>
        public int? displayIndex { get; set; }
        /// <summary>
        /// true: export, false: exclude from export
        /// </summary>
        public bool willExport { get; set; }

        /// <summary>
        /// used to replace the assessment prop (cell value) when a regex matches
        /// </summary>
        public CsvExportReplaceRuleObj ReplaceRulesObj { get; set; }
    }

    public class CsvExportReplaceRuleObj
    {
        public CsvExportReplaceRule[] ReplaceRules { get; set; }
        /// <summary>
        /// can be null
        /// </summary>
        public string ElseReplace { get; set; }
    }

    public class CsvExportReplaceRule
    {
        /// <summary>
        /// the regex text without js // and wihout flags
        /// </summary>
        public string Regex { get; set; }
        /// <summary>
        /// only multi line (m), ignore case (i)
        /// allow all possible permutations of i,g (checked with contains)
        /// </summary>
        public string RegexFlags { get; set; }
        /// <summary>
        /// if the regex matched then we replace the entire cell value with this value
        /// </summary>
        public string ReplaceWith { get; set; }
        /// <summary>
        /// this is set by the backend when we receive the regex and then convert it
        /// </summary>
        [JsonIgnore]
        public Regex RegexCompiled { get; set; }
    }



    /// <summary>
    /// used for download and upload of the csv assessments
    /// </summary>
    public class CsvExportImportAssessment
    {
        
        /// <summary>
        /// the user id (in case the tutor wants to see the solution NOT SUPPORTED YET)
        /// 
        /// PK
        /// </summary>
        public int UserId { get; set; }
        
        /// <summary>
        /// the token from the user (not for frontend)
        /// </summary>
        public string Token { get; set; }

        /// <summary>
        /// the name of the dir with the solution files
        /// </summary>
        public string SubmissionDirName { get; set; }

        /// <summary>
        /// the last name of the connected user
        /// </summary>
        public string LastName { get; set; }

        /// <summary>
        /// the first name of the connected user
        /// </summary>
        public string FirstName { get; set; }

        /// <summary>
        /// this is taken from the release participation so it counts for all p langs
        /// true: the student wants that his/her solution is not counted (for whatever), false: not
        /// can only be set once by the student
        /// </summary>
        public bool ShouldNotCount { get; set; }
       

        /// <summary>
        /// the used plang
        /// PK
        /// </summary>
        public int PLangId { get; set; }
        
        /// <summary>
        /// the used plang name
        /// </summary>
        public string PLangName { get; set; }

        /// <summary>
        /// a user can solve an exercise multiple times (connected through release.exercise)
        /// so this gives us the times the user participated on the exercise (through different releases)
        /// 
        /// IF a user solved the exercise with different p langs then this does not count separate because
        /// all restrictions are stored at the participation (e.g. time) and thus the p lang does not matter!
        /// 
        /// CAN be -1 on error
        /// 
        /// sometimes we want to know if a user solved an exercise multiple times
        /// e.g. when he/she is only allowed to solve it one time (e.g. an exam)
        /// </summary>
        public int ParticipationCount { get; set; }

        /// <summary>
        /// the summed points from the normal test
        /// </summary>
        public int? NormalTestPoints { get; set; }

        /// <summary>
        /// the summed max reachable normal test points 
        /// </summary>
        public int? MaxNormalTestPoints { get; set; }

        /// <summary>
        /// the summed test points from the submit tests
        /// </summary>
        public int? SubmitTestPoints { get; set; }

        /// <summary>
        /// the summed max reachable submit test points 
        /// </summary>
        public int? MaxSubmitTestPoints { get; set; }

        /// <summary>
        /// optional the manual given poitns (by the tutor)
        /// </summary>
        public int? ManualPoints { get; set; }

        /// <summary>
        /// the max manual points (just as a note), manual points can be more than this value
        /// </summary>
        public int? MaxManualPoints { get; set; }
        
        /// <summary>
        /// the note from the tutor
        /// </summary>
        public string NoteForOtherTutors { get; set; }

        /// <summary>
        /// the note for the student why he/she gets the manual points (feedback)
        /// </summary>
        public string FeedbackForStudent { get; set; }

        /// <summary>
        /// null or empty string for no error OR
        /// the error message from the test server or some other source
        /// </summary>
        public string LastAssessmentErrorMessage { get; set; }
    }

    /// <summary>
    /// these are the fields we really use form the csv import
    /// </summary>
    class InternalCsvImportAssessment
    {
        /// <summary>
        /// the user id (in case the tutor wants to see the solution NOT SUPPORTED YET)
        /// 
        /// PK
        /// </summary>
        public int UserId { get; set; }

        /// <summary>
        /// the used plang
        /// PK
        /// </summary>
        public int PLangId { get; set; }

        /// <summary>
        /// optional the manual given poitns (by the tutor)
        /// </summary>
        public int? ManualPoints { get; set; }

        /// <summary>
        /// the note from the tutor
        /// </summary>
        public string NoteForOtherTutors { get; set; }

        /// <summary>
        /// the note for the student why he/she gets the manual points (feedback)
        /// </summary>
        public string FeedbackForStudent { get; set; }
    }
}
