﻿using System;
using ClientServer.Db;
using System.Linq;
using System.Reactive.Linq;
using System.Threading.Tasks;
using ClientServer.Helpers;
using ClientServer.Models.Exercises.Release;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Schedulers
{
    /// <summary>
    /// a worker to check for exercise release changes
    /// we use a worker because we cannot start a (long living) thread in a controller ... the controller dies after the request is finished
    /// 
    /// it's also better in a worker because when we do this in a controller there might be 2 equal requets which would start release 2 times...
    /// </summary>
    public static class ReleaseScheduler
    {

        /// <summary>
        /// an object for thread locking (we actually don't know if locking works with asp core but we try...)
        /// </summary>
        private static readonly Object Locker = new object();

        /// <summary>
        /// starts periodically checking for releases to be released or unrelesed (and releases/unreleses them)
        /// </summary>
        public static void Start()
        {

            Observable
                .Interval(TimeSpan.FromMinutes(1))
                //.Interval(TimeSpan.FromSeconds(20)) //for debug
                .Subscribe((min) => { NextRun(); });
        }


        public static void NextRun()
        {

            //now = DateTime.SpecifyKind(now, DateTimeKind.Unspecified);

            //times in the db are stored as utc dateTimes
            var now = DateTime.Now.ToUniversalTime();

            if (AppConfiguration.IsDebugMode)
            {
                //TODO maybe enable??
                //Console.WriteLine("Release scheduler run # " + runNumber + " running at: " + now + " (UTC)");
            }

            var optionsBuilder = new DbContextOptionsBuilder<SyndromDbContext>();
            optionsBuilder.UseNpgsql(AppConfiguration.DbConnectionString);

            try
            {
                using (var context = new SyndromDbContext(optionsBuilder.Options))
                {
                    var automaticReleasesToRelease = context.ExerciseReleases
                        .Where(
                            p =>
                                p.ReleaseStartType == ReleaseStartType.Automatic
                                && p.AutomaticStartAt != null
                                && p.AutomaticStartAt.Value <= now
                                && p.ReleasedAt == null //after the release is unreleased (closed) the automatic would kick in because all is true 
                                //except releasedAt has now a value
                                && p.IsReleased == false
                                && p.Exercise.IsPermanentlyLocked == false)
                        .ToList()
                        ;

                    var automaticReleasesToUnrelease = context.ExerciseReleases
                        .Where(
                            p =>
                                p.ReleaseStartType == ReleaseStartType.Automatic
                                && p.ReleaseDurationType == ReleaseDurationType.Limited
                                && p.AutomaticEndAt != null
                                && p.AutomaticEndAt.Value <= now
                                && p.IsReleased
                                && p.Exercise.IsPermanentlyLocked == false)
                        .ToList()
                        ;

                    //release releases...
                    foreach (var exerciseRelease in automaticReleasesToRelease)
                    {
                        exerciseRelease.IsReleased = true;
                        exerciseRelease.ReleasedAt = now;
                        
                    }


                    foreach (var exerciseRelease in automaticReleasesToUnrelease)
                    {
                        exerciseRelease.IsReleased = false;

                        //assessment is now also done in a scheduler
//                        SubmissionAssessmentScheduler.StartNextNewThread(exerciseRelease.Id);
                    }

                    try
                    {
                        context.SaveChanges();
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.Message);
                    }

                    if (AppConfiguration.IsDebugMode)
                    {
                        //TODO maybe enable??
                        //Console.WriteLine("Release scheduler run finished, released: "
                        //                  + automaticReleasesToRelease.Count
                        //                  + ", unreleased: " + automaticReleasesToUnrelease.Count);
                    }
                }
            }
            catch (Exception exception)
            {
                Console.WriteLine(exception);
            }
        }
    }
}
